from typing import Optional, List, Union, Tuple

try:
    from typing import Unpack
except ImportError:
    from typing_extensions import Unpack


from string import ascii_lowercase, digits
from gdeltdoc.validation import validate_tone

Filter = Union[List[str], str]

VALID_TIMESPAN_UNITS = ["min", "h", "hours", "d", "days", "w", "weeks", "m", "months"]


def near(n: int, *args) -> str:
    """
    Build the filter to find articles containing words that occur within
    `n` words of each other.

    eg. near(5, "airline", "climate") finds articles containing both
    "airline" and "climate" within 5 words.
    """
    if len(args) < 2:
        raise ValueError("At least two words must be provided")

    return f"near{str(n)}:" + '"' + " ".join([a for a in args]) + '" '


def multi_near(
    nears: List[Tuple[int, Unpack[Tuple[str, ...]]]], method: str = "OR"
) -> str:
    """
    Build the filter to find articles containing multiple sets of near terms.

    eg. multi_near([(5, "airline", "crisis"), (10, "airline", "climate", "change")]) finds "airline" and "crisis"
    within 5 words, and/or "airline", "climate", and "change" within 10 words
    """
    if method not in ["AND", "OR"]:
        raise ValueError(f"method must be one of AND or OR, not {method}")

    formatted = [near(n, *args) for (n, *args) in nears]

    paren_flag = len(formatted) != 1 and method == "OR"
    l_pad, r_pad = paren_flag * "(", paren_flag * ") "

    return l_pad + f"{method} ".join(formatted) + r_pad


def repeat(n: int, keyword: str) -> str:
    """
    Build the filter to find articles containing `keyword` at least `n` times.

    eg. repeat(2, "environment") finds articles containing the word "environment"
    at least twice.

    Only single word repetitions are allowed.
    """
    if " " in keyword:
        raise ValueError("Only single words can be repeated")

    return f'repeat{str(n)}:"{keyword}" '


def multi_repeat(repeats: List[Tuple[int, str]], method: str) -> str:
    """
    Build the filter to find articles containing multiple repeated words using `repeat()`

    eg. multi_repeat([(2, "airline"), (3, "airport")], "AND") finds articles that contain the word "airline" at least
    twice and "airport" at least 3 times.

    Params
    ------
        repeats: A list of (int, str) tuples to be passed to `repeat()`. Eg. [(2, "airline"), (3, "airport")]
        method: How to combine the restrictions. Must be one of "AND" or "OR"
    """
    if method not in ["AND", "OR"]:
        raise ValueError(f"method must be one of AND or OR, not {method}")

    to_repeat = [repeat(n, keyword) for (n, keyword) in repeats]

    if method == "AND":
        return f"{method} ".join(to_repeat)
    else:
        # method == "OR"
        return "(" + f"{method} ".join(to_repeat) + ")"


class Filters:
    def __init__(
        self,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        timespan: Optional[str] = None,
        num_records: int = 250,
        keyword: Optional[Filter] = None,
        domain: Optional[Filter] = None,
        domain_exact: Optional[Filter] = None,
        near: Optional[str] = None,
        repeat: Optional[str] = None,
        country: Optional[Filter] = None,
        language: Optional[Filter] = None,
        theme: Optional[Filter] = None,
        tone: Optional[Filter] = None,
        tone_absolute: Optional[Filter] = None,
    ) -> None:
        """
        Construct filters for the GDELT API.

        Filters for `keyword`, `domain`, `domain_exact`, `country` and `theme`
        can be passed either as a single string or as a list of strings. If a list is
        passed, the values in the list are wrapped in a boolean OR.

        Params
        ------
        start_date
            The start date for the filter in YYYY-MM-DD format. The API officially only supports the
            most recent 3 months of articles. Making a request for an earlier date range may still
            return data, but it's not guaranteed.
            Must provide either `start_date` and `end_date` or `timespan`

        end_date
            The end date for the filter in YYYY-MM-DD format.

        timespan
            A timespan to search for, relative to the time of the request. Must match one of the API's timespan
            formats - https://blog.gdeltproject.org/gdelt-doc-2-0-api-debuts/
            Must provide either `start_date` and `end_date` or `timespan`

        num_records
            The number of records to return. Only used in article list mode and can be up to 250.

        keyword
            Return articles containing the exact phrase `keyword` within the article text.

        domain
            Return articles from the specified domain. Does not require an exact match so
            passing "cnn.com" will match articles from "cnn.com", "subdomain.cnn.com" and "notactuallycnn.com".

            NOTE: There's a bug in the API when filtering for domains that include a dash (-).
            If you supply a value for this filter using a domain with a dash you'll always get 0 results.

            As a workaround you can filter using the part of the domain after the dash (eg. cdn-apple.com -> apple.com).
            The query is an inexact match so this will return results from the domain you want, plus other domains
            which also match. You'll need to filter the results after you get them back to just the single domain.

            For more information on this bug, please see https://github.com/alex9smith/gdelt-doc-api/issues/35

        domain_exact
            Similar to `domain`, but requires an exact match.

        near
            Return articles containing words close to each other in the text. Use `near()` to construct.
            eg. near = near(5, "airline", "climate").

        repeat
            Return articles containing a single word repeated at least a number of times. Use `repeat()`
            to construct. eg. repeat = repeat(3, "environment").
            If you want to construct a filter with multiple repeated words, construct with `multi_repeat()`
            instead. eg. repeat = multi_repeat([(2, "airline"), (3, "airport")], "AND")

        country
            Return articles published in a country, formatted as the FIPS 2 letter country code.

        language
            Return articles published in the given language, formatted as the ISO 639 language code.

        theme
            Return articles that cover one of GDELT's GKG Themes. A full list of themes can be
            found here: http://data.gdeltproject.org/api/v2/guides/LOOKUP-GKGTHEMES.TXT

        tone
            Return articles above or below a particular tone score (ie more positive or more negative than
            a certain threshold). To use, specify either a greater than or less than sign and a positive
            or negative number (either an integer or floating point number). To find fairly positive
            articles, search for "tone>5" or to search for fairly negative articles, search for "tone<-5"

        tone_absolute
            The same as `tone` but ignores the positive/negative sign and lets you simply search for
            high emotion or low emotion articles, regardless of whether they were happy or sad in tone
        """
        self.query_params: List[str] = []
        self._valid_countries: List[str] = []
        self._valid_themes: List[str] = []

        # Check we have either start/end date or timespan, but not both
        if not start_date and not end_date and not timespan:
            raise ValueError("Must provide either start_date and end_date, or timespan")

        if start_date and end_date and timespan:
            raise ValueError(
                "Can only provide either start_date and end_date, or timespan"
            )

        if keyword:
            self.query_params.append(self._keyword_to_string(keyword))

        if domain:
            self.query_params.append(self._filter_to_string("domain", domain))

        if domain_exact:
            self.query_params.append(self._filter_to_string("domainis", domain_exact))

        if country:
            self.query_params.append(self._filter_to_string("sourcecountry", country))

        if language:
            self.query_params.append(self._filter_to_string("sourcelang", language))

        if theme:
            self.query_params.append(self._filter_to_string("theme", theme))

        if tone:
            validate_tone(tone)
            self.query_params.append(self._tone_to_string("tone", tone))

        if tone_absolute:
            validate_tone(tone_absolute)
            self.query_params.append(self._tone_to_string("toneabs", tone_absolute))

        if near:
            self.query_params.append(near)

        if repeat:
            self.query_params.append(repeat)

        if start_date:
            if not end_date:
                raise ValueError("Must provide both start_date and end_date")

            self.query_params.append(
                f'&startdatetime={start_date.replace("-", "")}000000'
            )
            self.query_params.append(f'&enddatetime={end_date.replace("-", "")}000000')

        elif timespan:
            self._validate_timespan(timespan)
            self.query_params.append(f"&timespan={timespan}")

        if num_records > 250:
            raise ValueError(f"num_records must 250 or less, not {num_records}")

        self.query_params.append(f"&maxrecords={str(num_records)}")

    @property
    def query_string(self) -> str:
        return "".join(self.query_params)

    @staticmethod
    def _filter_to_string(name: str, f: Filter) -> str:
        """
        Convert a Filter into the string representation needed for the API.

        Params
        ------
        name
            The filter name as required by the API

        f
            The Filter to convert

        Returns
        -------
        str
            The converted filter. Eg. "domain:cnn.com"
        """
        if type(f) == str:
            return f"{name}:{f} "

        else:
            # Build an OR statement
            return "(" + " OR ".join([f"{name}:{clause}" for clause in f]) + ") "

    @staticmethod
    def _keyword_to_string(keywords: Filter) -> str:
        """
        Convert a Filter for keywords into the string for the API.
        The keyword argument is different to all the others in that there's
        no parameter name needed and if a key phrase is passed it
        must be wrapped in quotes.

        Params
        ------
        keyword
            The keyword Filter

        Returns
        -------
        str
            The converted filter eg. "(airline OR shipping)"
        """
        if type(keywords) == str:
            return f'"{keywords}" '

        else:
            return (
                "("
                + " OR ".join(
                    [f'"{word}"' if " " in word else word for word in keywords]
                )
                + ") "
            )

    @staticmethod
    def _tone_to_string(name: str, tone: Filter) -> str:
        """
        Convert a Filter for tone or tone_absolute into the string for the API.
        Tone is different to the other parameters because it has no colon between
        the API parameter and the value.

        Params
        ------
        name
            The parameter name as required by the API. Must be either `tone` or `toneabs`
        tone
            The tone Filter

        Returns
        -------
        str
            The converted filter eg. "tone>5"
        """
        if type(tone) == str:
            return f"{name}{tone} "

        else:
            raise NotImplementedError("Multiple tone values are not supported yet.")

    @staticmethod
    def _validate_timespan(timespan: str) -> None:
        """
        Validate that the supplied timespan is in a format recognised by the API.
        Raises a `ValueError` if the timespan is not recognised.

        Supported timespan units are:
            - minutes - 15min
            - hours - 24h or 24hours
            - days - 30d or 30days
            - months - 2m or 2months

        Params
        ------
        timespan
            The timespan filter to be checked

        Returns
        -------
        None
        """

        value = timespan.rstrip(ascii_lowercase)
        unit = timespan[len(value) :]

        if unit not in VALID_TIMESPAN_UNITS:
            raise ValueError(
                f"Timespan {timespan} is invalid. {unit} is not a supported unit, must be one of {' '.join(VALID_TIMESPAN_UNITS)}"
            )

        if not all(d in digits for d in value):
            raise ValueError(
                f"Timespan {timespan} is invalid. {value} could not be converted into an integer"
            )

        if unit == "min" and int(value) < 60:
            raise ValueError(
                f"Timespan {timespan} is invalid. Period must be at least 60 minutes"
            )



from gdeltdoc import (
    Filters,
    near,
    multi_near,
    repeat,
    multi_repeat,
    VALID_TIMESPAN_UNITS,
)

import unittest


class FiltersTestCase(unittest.TestCase):
    """
    Test that the correct query strings are generated from
    various filters.
    """

    def test_single_keyword_filter(self):
        f = Filters(keyword="airline", start_date="2020-03-01", end_date="2020-03-02")
        self.assertEqual(
            f.query_string,
            '"airline" &startdatetime=20200301000000&enddatetime=20200302000000&maxrecords=250',
        )

    def test_single_keyphrase_filter(self):
        f = Filters(
            keyword="climate change", start_date="2020-03-01", end_date="2020-03-02"
        )
        self.assertEqual(
            f.query_string,
            '"climate change" &startdatetime=20200301000000&enddatetime=20200302000000&maxrecords=250',
        )

    def test_multiple_keywords(self):
        f = Filters(
            keyword=["airline", "climate"],
            start_date="2020-05-13",
            end_date="2020-05-14",
        )
        self.assertEqual(
            f.query_string,
            "(airline OR climate) &startdatetime=20200513000000&"
            "enddatetime=20200514000000&maxrecords=250",
        )

    def test_multiple_themes(self):
        f = Filters(
            theme=["ENV_CLIMATECHANGE", "LEADER"],
            start_date="2020-05-13",
            end_date="2020-05-14",
        )
        self.assertEqual(
            f.query_string,
            "(theme:ENV_CLIMATECHANGE OR theme:LEADER) &startdatetime=20200513000000&"
            "enddatetime=20200514000000&maxrecords=250",
        )

    def test_theme_and_keyword(self):
        f = Filters(
            keyword="airline",
            theme="ENV_CLIMATECHANGE",
            start_date="2020-05-13",
            end_date="2020-05-14",
        )
        self.assertEqual(
            f.query_string,
            '"airline" theme:ENV_CLIMATECHANGE &startdatetime=20200513000000&'
            "enddatetime=20200514000000&maxrecords=250",
        )

    def test_tone_filter(self):
        f = Filters(
            keyword="airline",
            start_date="2020-03-01",
            end_date="2020-03-02",
            tone=">10",
        )
        self.assertEqual(
            f.query_string,
            '"airline" tone>10 &startdatetime=20200301000000&enddatetime=20200302000000&maxrecords=250',
        )


class NearTestCast(unittest.TestCase):
    """
    Test that `near()` generates the right filters and errors.
    """

    def test_two_words(self):
        self.assertEqual(near(5, "airline", "crisis"), 'near5:"airline crisis" ')

    def test_three_words(self):
        self.assertEqual(
            near(10, "airline", "climate", "change"), 'near10:"airline climate change" '
        )

    def test_one_word(self):
        with self.assertRaisesRegex(ValueError, "At least two words"):
            near(5, "airline")


class MultiNearTestCast(unittest.TestCase):
    """
    Test that `near()` generates the right filters and errors.
    """

    def test_single_near(self):
        self.assertEqual(
            multi_near([(5, "airline", "crisis")]), near(5, "airline", "crisis")
        )

    def test_two_nears(self):
        self.assertEqual(
            multi_near(
                [(5, "airline", "crisis"), (10, "airline", "climate", "change")]
            ),
            "("
            + near(5, "airline", "crisis")
            + "OR "
            + near(10, "airline", "climate", "change")
            + ") ",
        )

    def test_two_nears_AND(self):
        self.assertEqual(
            multi_near(
                [(5, "airline", "crisis"), (10, "airline", "climate", "change")],
                method="AND",
            ),
            near(5, "airline", "crisis")
            + "AND "
            + near(10, "airline", "climate", "change"),
        )


class RepeatTestCase(unittest.TestCase):
    """
    Test that `repeat()` generates the correct filters and errors.
    """

    def test_repeat(self):
        self.assertEqual(repeat(3, "environment"), 'repeat3:"environment" ')

    def test_repeat_phrase(self):
        with self.assertRaisesRegex(ValueError, "single word"):
            repeat(5, "climate change   ")


class MultiRepeatTestCase(unittest.TestCase):
    """
    Test that `multi_repeat() generates the correct filters and errors.
    """

    def test_multi_repeat(self):
        self.assertEqual(
            multi_repeat([(2, "airline"), (3, "airport")], "AND"),
            'repeat2:"airline" AND repeat3:"airport" ',
        )

    def test_multi_repeat_or(self):
        self.assertEqual(
            multi_repeat([(2, "airline"), (3, "airport")], "OR"),
            '(repeat2:"airline" OR repeat3:"airport" )',
        )

    def test_multi_repeat_checks_method(self):
        with self.assertRaisesRegex(ValueError, "method must be one of AND or OR"):
            multi_repeat([(2, "airline"), (3, "airport")], "NOT_A_METHOD")


class TimespanTestCase(unittest.TestCase):
    """
    Test that `Filter._validate_timespan` validates timespans correctly
    """

    def test_allows_valid_units(self):
        for unit in VALID_TIMESPAN_UNITS:
            try:
                Filters._validate_timespan(f"60{unit}")
            except ValueError:
                self.fail()

    def test_forbids_invalid_units(self):
        with self.assertRaisesRegex(ValueError, "is not a supported unit"):
            Filters._validate_timespan(f"60milliseconds")

    def test_forbids_invalid_values(self):
        invalid_timespans = ["12.5min", "40days0", "2/3weeks"]
        for timespan in invalid_timespans:
            with self.assertRaises(ValueError):
                Filters._validate_timespan(timespan)

    def test_forbids_incorrectly_formatted_timespans(self):
        with self.assertRaisesRegex(ValueError, "is not a supported unit"):
            Filters._validate_timespan(f"min15")

    def test_timespan_greater_than_60_mins(self):
        with self.assertRaisesRegex(ValueError, "Period must be at least 60 minutes"):
            Filters._validate_timespan(f"15min")


class ToneFilterToStringTestCase(unittest.TestCase):
    def test_single_tone(self):
        self.assertEqual(Filters._tone_to_string("tone", ">5"), "tone>5 ")

    def test_rejects_multiple_tones(self):
        with self.assertRaisesRegex(
            NotImplementedError, "Multiple tone values are not supported yet"
        ):
            Filters._tone_to_string("tone", [">5", "<10"])



import json


def load_json(json_message, max_recursion_depth: int = 100, recursion_depth: int = 0):
    """
    tries to load a JSON formatted string and removes offending characters if present.

    :param json_message: The JSON string to load.
    :param max_recursion_depth: The maximum recursion depth allowed.
    :param recursion_depth: The current recursion depth.
    :return: The parsed JSON object.
    """
    try:
        if isinstance(json_message, bytes):
            json_message = json_message.decode()
        result = json.loads(json_message)

    except ValueError as e:
        if recursion_depth >= max_recursion_depth:
            raise ValueError("Max recursion depth is reached.")

        idx_to_replace = int(e.args[0].split(" ")[-1][:-1])
        json_message = (
            json_message[:idx_to_replace] + " " + json_message[idx_to_replace + 1 :]  # type: ignore
        )
        return load_json(json_message, max_recursion_depth, recursion_depth + 1)

    return result



import requests
import pandas as pd

from gdeltdoc.filters import Filters

from typing import Dict

from gdeltdoc.helpers import load_json

from gdeltdoc._version import version


class GdeltDoc:
    """
    API client for the GDELT 2.0 Doc API

    ```
    from gdeltdoc import GdeltDoc, Filters

    f = Filters(
        keyword = "climate change",
        start_date = "2020-05-10",
        end_date = "2020-05-11"
    )

    gd = GdeltDoc()

    # Search for articles matching the filters
    articles = gd.article_search(f)

    # Get a timeline of the number of articles matching the filters
    timeline = gd.timeline_search("timelinevol", f)
    ```

    ### Article List
    The article list mode of the API generates a list of news articles that match the filters.
    The client returns this as a pandas DataFrame with columns `url`, `url_mobile`, `title`,
    `seendate`, `socialimage`, `domain`, `language`, `sourcecountry`.

    ### Timeline Search
    There are 5 available modes when making a timeline search:
    * `timelinevol` - a timeline of the volume of news coverage matching the filters,
        represented as a percentage of the total news articles monitored by GDELT.
    * `timelinevolraw` - similar to `timelinevol`, but has the actual number of articles
        and a total rather than a percentage
    * `timelinelang` - similar to `timelinevol` but breaks the total articles down by published language.
        Each language is returned as a separate column in the DataFrame.
    * `timelinesourcecountry` - similar to `timelinevol` but breaks the total articles down by the country
        they were published in. Each country is returned as a separate column in the DataFrame.
    * `timelinetone` - a timeline of the average tone of the news coverage matching the filters.
        See [GDELT's documentation](https://blog.gdeltproject.org/gdelt-doc-2-0-api-debuts/)
        for more information about the tone metric.
    """

    def __init__(self, json_parsing_max_depth: int = 100) -> None:
        """
        Params
        ------
        json_parsing_max_depth
            A parameter for the json parsing function that removes illegal character. If 100 it will remove at max
            100 characters before exiting with an exception
        """
        self.max_depth_json_parsing = json_parsing_max_depth

    def article_search(self, filters: Filters) -> pd.DataFrame:
        """
        Make a query against the `ArtList` API to return a DataFrame of news articles that
        match the supplied filters.

        Params
        ------
        filters
            A `gdelt-doc.Filters` object containing the filter parameters for this query.

        Returns
        -------
        pd.DataFrame
            A pandas DataFrame of the articles returned from the API.
        """
        articles = self._query("artlist", filters.query_string)
        if "articles" in articles:
            return pd.DataFrame(articles["articles"])
        else:
            return pd.DataFrame()

    def timeline_search(self, mode: str, filters: Filters) -> pd.DataFrame:
        """
        Make a query using one of the API's timeline modes.

        Params
        ------
        mode
            The API mode to call. Must be one of "timelinevol", "timelinevolraw",
            "timelinetone", "timelinelang", "timelinesourcecountry".

            See https://blog.gdeltproject.org/gdelt-doc-2-0-api-debuts/ for a
            longer description of each mode.

        filters
            A `gdelt-doc.Filters` object containing the filter parameters for this query.

        Returns
        -------
        pd.DataFrame
            A pandas DataFrame of the articles returned from the API.
        """
        timeline = self._query(mode, filters.query_string)

        # If no results
        if (timeline == {}) or (len(timeline["timeline"]) == 0):
            return pd.DataFrame()

        results = {
            "datetime": [entry["date"] for entry in timeline["timeline"][0]["data"]]
        }

        for series in timeline["timeline"]:
            results[series["series"]] = [entry["value"] for entry in series["data"]]

        if mode == "timelinevolraw":
            results["All Articles"] = [
                entry["norm"] for entry in timeline["timeline"][0]["data"]
            ]

        formatted = pd.DataFrame(results)
        formatted["datetime"] = pd.to_datetime(formatted["datetime"])

        return formatted

    def _query(self, mode: str, query_string: str) -> Dict:
        """
        Submit a query to the GDELT API and return the results as a parsed JSON object.

        Params
        ------
        mode
            The API mode to call. Must be one of "artlist", "timelinevol",
            "timelinevolraw", "timelinetone", "timelinelang", "timelinesourcecountry".

        query_string
            The query parameters and date range to call the API with.

        Returns
        -------
        Dict
            The parsed JSON response from the API.
        """
        if mode not in [
            "artlist",
            "timelinevol",
            "timelinevolraw",
            "timelinetone",
            "timelinelang",
            "timelinesourcecountry",
        ]:
            raise ValueError(f"Mode {mode} not in supported API modes")

        headers = {
            "User-Agent": f"GDELT DOC Python API client {version} - https://github.com/alex9smith/gdelt-doc-api"
        }

        response = requests.get(
            f"https://api.gdeltproject.org/api/v2/doc/doc?query={query_string}&mode={mode}&format=json",
            headers=headers,
        )

        if response.status_code not in [200, 202]:
            raise ValueError(
                "The gdelt api returned a non-successful statuscode. This is the response message: {}".format(
                    response.text
                )
            )

        # Response is text/html if it's an error and application/json if it's ok
        if "text/html" in response.headers["content-type"]:
            raise ValueError(
                f"The query was not valid. The API error message was: {response.text.strip()}"
            )

        return load_json(response.content, self.max_depth_json_parsing)



from gdeltdoc.validation import validate_tone
import unittest


class ValidateToneTestCase(unittest.TestCase):
    def valid_tone_doesnt_raise_error(self):
        validate_tone(">5")

    def raises_when_comparator_missing(self):
        with self.assertRaises(ValueError):
            validate_tone("10")

    def raises_when_equals_in_comparator(self):
        with self.assertRaises(ValueError):
            validate_tone(">=10")

    def raises_when_multiple_tones(self):
        with self.assertRaises(NotImplementedError):
            validate_tone([">5", "<10"])



import unittest.mock
import pandas as pd
import unittest

from gdeltdoc import GdeltDoc, Filters
from datetime import datetime, timedelta


class ArticleSearchTestCast(unittest.TestCase):
    """
    Test that the API client behaves correctly when doing an article search query
    """

    def setUp(self):
        self.start_date = (datetime.today() - timedelta(days=7)).strftime("%Y-%m-%d")
        self.end_date = (datetime.today() - timedelta(days=6)).strftime("%Y-%m-%d")

        f = Filters(
            keyword="environment", start_date=self.start_date, end_date=self.end_date
        )
        self.articles = GdeltDoc().article_search(f)

    def tearDown(self):
        pass

    def test_articles_is_a_df(self):
        self.assertEqual(type(self.articles), pd.DataFrame)

    def test_correct_columns(self):
        self.assertEqual(
            list(self.articles.columns),
            [
                "url",
                "url_mobile",
                "title",
                "seendate",
                "socialimage",
                "domain",
                "language",
                "sourcecountry",
            ],
        )

    def test_rows_returned(self):
        # This test could fail if there really are no articles
        # that match the filter, but given the query used for
        # This tests could fail if there really are no articles
        # that match the filter, but given the query used for
        # testing that's very unlikely.
        self.assertGreaterEqual(self.articles.shape[0], 1)


class TimelineSearchTestCase(unittest.TestCase):
    """
    Test that the various modes of timeline search behave correctly.
    """

    # Make one set of API calls per tests suite run, not one per test
    @classmethod
    def setUpClass(cls):
        cls.start_date = (datetime.today() - timedelta(days=7)).strftime("%Y-%m-%d")
        cls.end_date = (datetime.today() - timedelta(days=6)).strftime("%Y-%m-%d")

        f = Filters(
            keyword="environment", start_date=cls.start_date, end_date=cls.end_date
        )

        gd = GdeltDoc()
        cls.all_results = [
            gd.timeline_search(mode, f)
            for mode in [
                "timelinevol",
                "timelinevolraw",
                "timelinelang",
                "timelinetone",
                "timelinesourcecountry",
            ]
        ]

    def test_all_modes_return_a_df(self):
        self.assertTrue(
            all([type(result) == pd.DataFrame for result in self.all_results])
        )

    def test_all_modes_return_data(self):
        self.assertTrue(all([result.shape[0] >= 1 for result in self.all_results]))

    def test_unsupported_mode(self):
        with self.assertRaisesRegex(ValueError, "not in supported API modes"):
            GdeltDoc().timeline_search(
                "unsupported",
                Filters(
                    keyword="environment",
                    start_date=self.start_date,
                    end_date=self.end_date,
                ),
            )

    def test_vol_has_two_columns(self):
        self.assertEqual(self.all_results[0].shape[1], 2)

    def test_vol_raw_has_three_columns(self):
        self.assertEqual(self.all_results[1].shape[1], 3)

    def test_handles_empty_API_response(self):
        gd = GdeltDoc()
        with unittest.mock.patch.object(gd, "_query") as query_mock:
            query_mock.return_value = {}
            result = gd.timeline_search(
                "timelinetone", Filters(keyword="environment", timespan="1h")
            )
            self.assertTrue(type(result) == pd.DataFrame)
            self.assertEqual(result.shape[0], 0)


class QueryTestCase(unittest.TestCase):

    def test_handles_invalid_query_string(self):
        with self.assertRaisesRegex(
            ValueError, "The query was not valid. The API error message was"
        ):
            GdeltDoc()._query("artlist", "environment&timespan=mins15")



from typing import List, Union

Filter = Union[List[str], str]


def validate_tone(tone: Filter) -> None:
    if not ("<" in tone or ">" in tone):
        raise ValueError("Tone must contain either greater than or less than")

    if "=" in tone:
        raise ValueError("Tone cannot contain '='")

    if type(tone) == list:
        raise NotImplementedError("Multiple tones are not supported yet")



import setuptools

with open("requirements.txt", "r") as f:
    requirements = [line.replace("\n", "") for line in f.readlines()]

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("gdeltdoc/_version.py", "r") as g:
    version = "1.0.0"
    for line in g.readlines():
        if "version" in line:
            version = (
                line.split("=")[1].replace("\n", "").replace('"', "").replace(" ", "")
            )

setuptools.setup(
    name="gdeltdoc",
    version=version,
    author="Alex Whitehead-Smith",
    author_email="alex.me.smith@gmail.com",
    description="A client for the GDELT 2.0 Doc API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/alex9smith/gdelt-doc-api",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    install_requires=requirements,
)



from gdeltdoc.api_client import GdeltDoc
from gdeltdoc.filters import Filters, near, multi_near, repeat, multi_repeat, VALID_TIMESPAN_UNITS
from gdeltdoc._version import version

__version__ = version
